__all__ = ['factory']

import types
from edge import eventlog
from .exceptions import PsuCtlError


class _LedUnit10(types.IUnit):
    _ACK_TOUT_SEC = 1

    _REF_TO_OPREG_MAP = {
        ('led', 'LED_L1RG'): 0x0000,
        ('led', 'LED_L2RG'): 0x0001,
        ('led', 'LED_L3RG'): 0x0002,
        ('led', 'LED_L4RG'): 0x0003,
        ('led', 'LED_L5RG'): 0x0004
    }

    _MODE_TO_OPVAL_MAP = {
        (None, None): 0x0000,
        ('G', '0HZ'): 0x0001,
        ('R', '0HZ'): 0x0002,
        ('Y', '0HZ'): 0x0003,

        ('G', '1HZ'): 0x0101,
        ('R', '1HZ'): 0x0102,
        ('Y', '1HZ'): 0x0103,

        ('G', '2HZ'): 0x0201,
        ('R', '2HZ'): 0x0202,
        ('Y', '2HZ'): 0x0203,

        ('GR', '1HZ'): 0x0112,
        ('GY', '1HZ'): 0x0113,
        ('YR', '1HZ'): 0x0123,

        ('GR', '2HZ'): 0x0212,
        ('GY', '2HZ'): 0x0213,
        ('YR', '2HZ'): 0x0223
    }

    def __init__(self, firmware, ref):
        """
        :type firmware: types.IFirmware
        :type ref: list | tuple(str, str)
        """
        if tuple(ref) not in self._REF_TO_OPREG_MAP:
            eventlog.log_e('fail: ivalid ref <{}>'.format(ref))
            raise PsuCtlError

        types.IUnit.__init__(self, firmware, ref)

    def run(self, conf):
        """
        :param conf: psuctl.conf#/scripts/<name>/conf
        :type conf: dict
        :rtype: None
        """
        color = conf.get('color')
        if color and type(color) not in [str, unicode]:
            eventlog.log_e_param('#/color')
            raise PsuCtlError

        blink = conf.get('blink')
        if blink and type(blink) not in [str, unicode]:
            eventlog.log_e_param('#/blink')
            raise PsuCtlError

        mode = (color, blink)
        op_val = self._MODE_TO_OPVAL_MAP.get(tuple(mode))
        if op_val is None:
            eventlog.log_e_param('#/mode')
            raise PsuCtlError

        op_reg = self._REF_TO_OPREG_MAP[tuple(self._ref)]
        try:
            answer = self._firmware.write_rw_reg(op_reg, op_val, self._ACK_TOUT_SEC)
            if op_val != answer:
                eventlog.log_w('fail: write_rw_reg (query:{}, reply:{})'.format(op_val, answer))
        except PsuCtlError:
            eventlog.log_e('fail: write_rw_reg')
            raise


def factory(firmware, unit_ref=None):
    """
    :type firmware: types.IFirmware
    :type unit_ref: list | tuple(str, str)
    :rtype: IUnit | None
    """
    firmware_ver = firmware.version
    if firmware_ver in ['v1.0', 'v1.1', 'v1.2', 'v1.3', 'v1.4']:
        unit_factory = _LedUnit10
    else:
        unit_factory = _LedUnit10
        eventlog.log_w('fail: unsupported firmware <{}> (using {})'
                       .format(firmware_ver, unit_factory.__name__))

    return unit_factory(firmware, unit_ref)
