__all__ = ['sh_unit_factory', 'sh_ics_unit_factory']

import types
from edge import eventlog
from .exceptions import PsuCtlError

from edge.com import sh


class _ShUnit(types.IUnit):
    def __init__(self, firmware, ref):
        """
        :type firmware: types.IFirmware | types.IFirmware2 | None
        :type ref: list | tuple(str, str)
        """
        types.IUnit.__init__(self, firmware, ref)

    def run(self, conf):
        """
        :param conf: psuctl.conf#/scripts/<name>/conf
        :type conf: dict
        :rtype: None
        """
        shell_cmd = conf.get('shell')
        if shell_cmd and type(shell_cmd) not in [str, unicode]:
            eventlog.log_e_param('#/shell')
            raise PsuCtlError

        try:
            rc = sh.run_silent(shell_cmd)
            if rc == 0:
                eventlog.log_d('done: {}'.format(shell_cmd))
            else:
                eventlog.log_d('fail: {} rc {}'.format(shell_cmd, rc))

        except Exception:
            eventlog.log_x()
            raise PsuCtlError


class _ShIcsUnit(types.IUnit):
    def __init__(self, firmware, ref):
        """
        :type firmware: types.IFirmware | types.IFirmware2 | None
        :type ref: list | tuple(str, str)
        """
        types.IUnit.__init__(self, firmware, ref)

    def run(self, conf):
        """
        :param conf: psuctl.conf#/scripts/<name>/conf
        :type conf: dict
        :rtype: None
        """
        action = conf.get('action')

        if action not in ['rotate', 'get-limits']:
            eventlog.log_e_param('#/action')
            raise PsuCtlError

        if action == 'get-limits':
            ch_max = conf.get('max-steps')
            ch_min = conf.get('min-steps', 0)
            if ch_max is None:
                eventlog.log_e('fail:#/max-steps')
                raise PsuCtlError
            return ch_min, ch_max

        raw_steps = conf.get('steps')
        if type(raw_steps) in [str, unicode]:
            raw_steps = int(raw_steps)
        elif type(raw_steps) in [int]:
            pass
        else:
            eventlog.log_e_param('#/steps')
            raise PsuCtlError

        rot_steps = abs(raw_steps)

        shell_cmd = conf.get('shell')
        if shell_cmd and type(shell_cmd) not in [str, unicode]:
            eventlog.log_e_param('#/shell')
            raise PsuCtlError

        shell_cmd += ' {}'.format(raw_steps)
        try:
            rc = sh.run_silent(shell_cmd)
            if rc == 0:
                eventlog.log_d('done: {}'.format(shell_cmd))
            else:
                eventlog.log_d('fail: {} rc {}'.format(shell_cmd, rc))
                raise PsuCtlError

        except Exception:
            eventlog.log_x()
            raise PsuCtlError


def sh_unit_factory(firmware, unit_ref=None):
    """
    :type firmware: types.IFirmware
    :type unit_ref: list | tuple(str, str)
    :rtype: IUnit | None
    """
    unit_factory = _ShUnit
    return unit_factory(firmware, unit_ref)


def sh_ics_unit_factory(firmware, unit_ref=None):
    """
    :type firmware: types.IFirmware
    :type unit_ref: list | tuple(str, str)
    :rtype: IUnit | None
    """
    unit_factory = _ShIcsUnit
    return unit_factory(firmware, unit_ref)
