__all__ = ['factory']

import time
from edge import eventlog
from .types import IUnit
from .exceptions import PsuCtlError


_ROT_DIR_CW = 0x00
_ROT_DIR_CCW = 0x100
_MIN_ACK_TOUT_SEC = 0.25


class _StepperUnit10(IUnit):
    _MAX_STEP_COUNT = 0x7FFF
    _MAX_OP_STEP_COUNT = 0x00FF

    _REF_TO_INFO_MAP = {
        ('stepper', 'BASE_1-4S2M_MOTOR1'): {
            'reg': 0x0011,
        },
        ('stepper', 'BASE_1-4S2M_MOTOR2'): {
            'reg': 0x0012,
        },
        ('stepper', 'BASE_2-4S2M_MOTOR1'): {
            'reg': 0x0021,
        },
        ('stepper', 'BASE_2-4S2M_MOTOR2'): {
            'reg': 0x0022,
        }
    }

    def __init__(self, firmware, ref):
        """
        :type firmware: types.IFirmware
        :type ref: list | tuple(str, str)
        """
        if tuple(ref) not in self._REF_TO_INFO_MAP:
            eventlog.log_e('fail: ivalid ref <{}>'.format(ref))
            raise PsuCtlError

        IUnit.__init__(self, firmware, ref)

    def run(self, conf):
        """
        :param conf: psuctl.conf#/scripts/<name>/conf
        :type conf: dict
        :rtype: None | (int, int)
        """
        # unpack #/action
        action = conf.get('action')

        if action not in ['rotate', 'get-limits']:
            eventlog.log_e_param('#/action')
            raise PsuCtlError

        if action == 'get-limits':
            ch_min = conf.get('min-steps', 0)
            ch_max = conf.get('max-steps')
            if ch_max is None:
                eventlog.log_e('fail:#/max-steps')
                raise PsuCtlError
            return ch_min, ch_max

        raw_steps = conf.get('steps')
        if type(raw_steps) in [str, unicode]:
            raw_steps = int(raw_steps)
        elif type(raw_steps) in [int]:
            pass
        else:
            eventlog.log_e_param('#/steps')
            raise PsuCtlError

        rot_steps = abs(raw_steps)
        if rot_steps > self._MAX_STEP_COUNT:
            eventlog.log_e('fail: invalid steps value <{}>'.format(raw_steps))
            raise PsuCtlError

        rot_dir = _ROT_DIR_CW if raw_steps > 0 else _ROT_DIR_CCW

        max_count = int(raw_steps / self._MAX_OP_STEP_COUNT)
        addition_step = int(raw_steps % self._MAX_OP_STEP_COUNT)
        count = 0
        while count < max_count:
            self._rotate255(self._MAX_OP_STEP_COUNT, rot_dir)
            # A time required to switch board's RS485 from rx to tx (or vice versa).
            time.sleep(0.025)
            count += 1

        self._rotate255(addition_step, rot_dir)

    def _rotate255(self, rot_steps, rot_dir):
        """
        :param rot_steps: value [0, 255]
        :type rot_steps: int
        :type rot_dir: int
        :rtype: None
        """
        op_tout_sec = _MIN_ACK_TOUT_SEC + 3 * rot_steps * 0.003 * 2

        op_reg = self._REF_TO_INFO_MAP[tuple(self._ref)]['reg']
        op_val = rot_steps + rot_dir

        try:
            answer = self._firmware.write_rw_reg(op_reg, op_val, op_tout_sec)
            if op_val != answer:
                eventlog.log_w('fail: write_rw_reg (query:{}, reply:{})'.format(op_val, answer))
        except PsuCtlError:
            eventlog.log_e('fail: write_rw_reg')
            raise


class _StepperUnit11(IUnit):
    _MAX_STEP_COUNT = 0x7FFF
    _REF_TO_INFO_MAP = {
        ('stepper', 'BASE_1-4S2M_MOTOR1'): {
            'slow': 0x0011,
            'fast': 0x0031
        },
        ('stepper', 'BASE_1-4S2M_MOTOR2'): {
            'slow': 0x0012,
            'fast': 0x0032
        },
        ('stepper', 'BASE_2-4S2M_MOTOR1'): {
            'slow': 0x0021,
            'fast': 0x0041
        },
        ('stepper', 'BASE_2-4S2M_MOTOR2'): {
            'slow': 0x0022,
            'fast': 0x0042
        }
    }

    def __init__(self, firmware, ref):
        """
        :type firmware: types.IFirmware
        :type ref: list | tuple(str, str)
        """
        if tuple(ref) not in self._REF_TO_INFO_MAP:
            eventlog.log_e('fail: ivalid ref <{}>'.format(ref))
            raise PsuCtlError

        IUnit.__init__(self, firmware, ref)

    def run(self, conf):
        """
        :param conf: psuctl.conf#/scripts/<name>/conf
        :type conf: dict
        :rtype: None | (int, int)
        """
        # unpack #/action
        action = conf.get('action')

        if action not in ['rotate', 'get-limits']:
            eventlog.log_e_param('#/action')
            raise PsuCtlError

        if action == 'get-limits':
            min_steps = conf.get('min-steps', 0)
            max_steps = conf.get('max-steps')
            if max_steps is None:
                eventlog.log_e('fail:#/max-steps')
                raise PsuCtlError
            return min_steps, max_steps

        # unpack #/steps
        raw_steps = conf.get('steps')
        if type(raw_steps) in [str, unicode]:
            raw_steps = int(raw_steps)
        elif type(raw_steps) in [int]:
            pass
        else:
            eventlog.log_e_param('#/steps')
            raise PsuCtlError

        rot_steps = abs(raw_steps)
        if rot_steps > self._MAX_STEP_COUNT:
            eventlog.log_e('fail: invalid steps value <{}>'.format(raw_steps))
            raise PsuCtlError

        rot_dir = _ROT_DIR_CW if raw_steps > 0 else _ROT_DIR_CCW

        # unpack #/speed
        raw_speed = conf.get('speed', 'slow')
        if raw_speed == 'slow':
            op_tout_sec = _MIN_ACK_TOUT_SEC + 3 * rot_steps * 0.003 * 2

        elif raw_speed == 'fast':
            op_tout_sec = _MIN_ACK_TOUT_SEC + rot_steps * 0.003

        else:
            eventlog.log_e_param('#/speed')
            raise PsuCtlError

        # execute operation
        op_reg = self._REF_TO_INFO_MAP[tuple(self._ref)][raw_speed]
        op_val = int(rot_steps / 256) * 0x200 + (rot_steps % 256) + rot_dir
        try:
            answer = self._firmware.write_rw_reg(op_reg, op_val, op_tout_sec)
            if op_val != answer:
                eventlog.log_w('fail: write_rw_reg (query:{}, reply:{})'.format(op_val, answer))
        except PsuCtlError:
            eventlog.log_e('fail: write_rw_reg')
            raise


class _StepperUnit13(IUnit):
    _MAX_STEP_COUNT = 0xFFFF
    _REF_TO_INFO_MAP = {
        ('stepper', 'BASE_1-4S2M_MOTOR1'): {
            'reg': 0x0011,
        },
        ('stepper', 'BASE_1-4S2M_MOTOR1', 'LIMITER'): {
            'reg': 0x0111
        },
        ('stepper', 'BASE_1-4S2M_MOTOR2'): {
            'reg': 0x0012
        },
        ('stepper', 'BASE_1-4S2M_MOTOR2', 'LIMITER'): {
            'reg': 0x0112
        },
        ('stepper', 'BASE_2-4S2M_MOTOR1'): {
            'reg': 0x0021
        },
        ('stepper', 'BASE_2-4S2M_MOTOR1', 'LIMITER'): {
            'reg': 0x0121
        },
        ('stepper', 'BASE_2-4S2M_MOTOR2'): {
            'reg': 0x0022
        },
        ('stepper', 'BASE_2-4S2M_MOTOR2', 'LIMITER'): {
            'reg': 0x0122
        }
    }

    def __init__(self, firmware, ref):
        """
        :type firmware: types.IFirmware
        :type ref: list
        """
        if tuple(ref) not in self._REF_TO_INFO_MAP:
            eventlog.log_e('fail: ivalid ref <{}>'.format(ref))
            raise PsuCtlError

        IUnit.__init__(self, firmware, ref)

    def run(self, conf):
        """
        :param conf: psuctl.conf#/scripts/<name>/conf
        :type conf: dict
        :rtype: None | (int, int)
        """
        # unpack #/action
        action = conf.get('action')

        if action not in ['rotate', 'get-limits']:
            eventlog.log_e_param('#/action')
            raise PsuCtlError

        if action == 'get-limits':
            min_steps = conf.get('min-steps', 0)
            max_steps = conf.get('max-steps')
            if max_steps is None:
                eventlog.log_e('fail:#/max-steps')
                raise PsuCtlError

            return min_steps, max_steps

        # unpack #/steps
        raw_steps = conf.get('steps')
        if type(raw_steps) in [str, unicode]:
            raw_steps = int(raw_steps)

        elif type(raw_steps) in [int]:
            pass

        else:
            eventlog.log_e_param('#/steps')
            raise PsuCtlError

        rot_steps = abs(raw_steps)
        if rot_steps > self._MAX_STEP_COUNT:
            eventlog.log_e('fail: invalid steps value <{}>'.format(raw_steps))
            raise PsuCtlError

        rot_dir = _ROT_DIR_CW if raw_steps > 0 else _ROT_DIR_CCW

        # unpack #/speed
        raw_speed = conf.get('speed', 0)
        if raw_speed == 'slow':
            rot_speed = 63

        elif raw_speed == 'fast':
            rot_speed = 0

        else:
            raw_speed = int(raw_speed)

            if 0 <= raw_speed <= 63:
                rot_speed = raw_speed

            else:
                eventlog.log_e_param('#/speed')
                raise PsuCtlError

        # execute operation
        if 'LIMITER' in self._ref:
            op_tout_sec = 2 + 2 * rot_steps * 0.003 * (rot_speed + 1)

        else:
            op_tout_sec = _MIN_ACK_TOUT_SEC + rot_steps * 0.003 * (rot_speed + 1)

        def pack_op_value():
            """
            :return: Packed steps value to send to psu-ctl register , real steps count
            :rtype: (int, int)
            """
            if rot_steps > 255:
                power_val = 0
                steps_val = int(rot_steps / 2)
                while steps_val > 0:
                    power_val += 1
                    steps_val = int(steps_val / 2)
                value = rot_dir + 0x8000 + power_val
                steps = 2 ** power_val

            else:
                value = rot_dir + rot_steps
                steps = rot_steps

            value += rot_speed * 0x200
            return value, steps

        def _try_rotate():
            op_reg = self._REF_TO_INFO_MAP[tuple(self._ref)]['reg']
            try:
                eventlog.log_d("rotate {} steps, send {}".format(real_steps, op_val))
                answer = self._firmware.write_rw_reg(op_reg, op_val, op_tout_sec)
                if op_val != answer:
                    eventlog.log_w('fail: write_rw_reg (query:{}, reply:{})'.format(op_val, answer))

            except PsuCtlError:
                eventlog.log_e('fail: write_rw_reg')
                raise

        op_val, real_steps = pack_op_value()
        while real_steps < rot_steps:

            _try_rotate()
            rot_steps = rot_steps - real_steps
            op_val, real_steps = pack_op_value()

        op_val, real_steps = pack_op_value()
        _try_rotate()


def factory(firmware, unit_ref):
    """
    :type firmware: types.IFirmware
    :type unit_ref: list | tuple(str, str)
    :rtype: IUnit | None
    """
    firmware_ver = firmware.version
    if firmware_ver in ['v1.0']:
        unit_factory = _StepperUnit10

    elif firmware_ver in ['v1.1', 'v1.2']:
        unit_factory = _StepperUnit11

    elif firmware_ver in ['v1.3', 'v1.4', 'v1.5']:
        unit_factory = _StepperUnit13

    else:
        unit_factory = _StepperUnit13
        eventlog.log_w('fail: unsupported firmware <{}> (using {})'
                       .format(firmware_ver, unit_factory.__name__))

    return unit_factory(firmware, unit_ref)
